#include "QuadTreeRenderer.h"

#include <GL/gl.h>
#include <GL/glu.h>
#include "Assert.h"
#include "Vector.h"
#include "glFuncs.h"
#include "loadbmp.h"

#include <math.h> /* for fabs() */
#include <stdlib.h>

QuadTreeRenderer::QuadTreeRenderer()
{
	map = NULL;
	QuadTreeRender = NULL;

	index_data = NULL;
	index_pointer = NULL;
	multi_indexes = NULL;
	multi_lengths = NULL;
	multi_indexes_ARB = NULL;

	quadtree_data = NULL;
	vertex_data = NULL;
	color_data = NULL;
	texture_coords = NULL;

	d_array = NULL;

	compiled_array = false;
	multi_draw = false;

	text_data = NULL;
	text_width = text_height = 0;
}

bool QuadTreeRenderer::load_texture(char *name)
{
	BITMAP bm;
	int i;
	GLint max_size;

	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max_size);

	if (LoadTextureFromBitmapFile(name, &bm) == NULL)
	{
		return false;
	}

	text_width	= bm.bmWidth;
	text_height	= bm.bmHeight;
	text_data   = bm.bmBits;

	if (text_width > max_size)
	{
		WARN("texture width exceeds max_size --> disabled");
		return false;
	}

	if (text_height > max_size)
	{
		WARN("texture width exceeds max_size --> disabled");
		return false;
	}

	for (i = 2; i < 20; i++)
	{
		if (pow(2, i) == text_width) break;
	}
	if (i == 20)
	{
		WARN("texture bitmap hast to be of edge len 2**n!");
	}

	return true;
}

GEZERROR QuadTreeRenderer::Initialize(Console &cons, Preferences &params)
{
	/* initialize opengl, etc... */
	unsigned long i, j, k;

	ASSERT(map != NULL, "can't initialize renderer without loaded map!");
	if (map == NULL) return GEZ_UNKNOWN;
	
	ASSERT(console == NULL, "console not null");

	SetConsole(&cons);

	console->Separator(true);

#define RESERVE_MEM(name, type, dimension)\
	if (name != NULL)\
	{\
		delete[] name;\
		name = NULL;\
	}\
	name = (type *)new type[dimension];\
	ASSERT(name != NULL, "kein platz mehr da");\
	if (name == NULL)\
	{\
		CleanUp();\
		return GEZ_UNKNOWN;\
	}

	MAP_WIDTH = map->width;
	MAP_HEIGHT = map->height;
	
	RESERVE_MEM(quadtree_data, GLbyte, MAP_WIDTH * MAP_HEIGHT);
	RESERVE_MEM(vertex_data, vertex, MAP_WIDTH * MAP_HEIGHT);
	RESERVE_MEM(color_data, vertex, MAP_WIDTH * MAP_HEIGHT);
	RESERVE_MEM(texture_coords, text_coord, MAP_WIDTH * MAP_HEIGHT);
	RESERVE_MEM(d_array, GLfloat, MAP_WIDTH * MAP_HEIGHT);

	console->Output("Setting preferences... \n");

	/* get preferences */
	C = params.GET_FLOAT_PARAM(PARAM_GLOBAL_DETAIL);
	console->Output("'C' (Global Detail) is %f\n", C);

	c = params.GET_FLOAT_PARAM(PARAM_LOCAL_DETAIL);
	console->Output("'c' (Local Detail) is %f\n", c);

	min_detail = params.GET_INT_PARAM(PARAM_MIN_DETAIL);
	console->Output("Minimum Detail is %d\n", min_detail);

	max_detail = params.GET_INT_PARAM(PARAM_MAX_DETAIL);
	console->Output("Maximum Detail is %d\n", max_detail);

	frustum_culling = params.GET_INT_PARAM(PARAM_FRUSTUM_CULLING) != 0;
	console->Output("frustum culling is %s\n", frustum_culling? "enabled" : "disabled");

	show_cull = params.GET_INT_PARAM(PARAM_SHOW_CULL) != 0;

	console->Separator();
	console->Output("Initializing QuadTree structure... ");

	/* prepare vertex data*/
	texture_mapping = false;
	if (params.GET_INT_PARAM(PARAM_TEXTURE_MAPPING))
	{
		if (!(texture_mapping = load_texture(params.GET_STRING_PARAM(PARAM_TEXTURE_NAME))))
		{
			WARN("error loading texture");
		}
	}

	float TEXT_EDGE_LEN_X;
	float TEXT_EDGE_LEN_Y;

	TEXT_EDGE_LEN_X = params.GET_FLOAT_PARAM(PARAM_TEXTURE_EDGE_LEN);

	// if text scale is 0 the map the texture is mapped to the complete map
	if (TEXT_EDGE_LEN_X < 1)
	{
		TEXT_EDGE_LEN_X = map->width;
		TEXT_EDGE_LEN_Y = map->height;
	}
	else
	{
		TEXT_EDGE_LEN_Y = TEXT_EDGE_LEN_X * ( float(text_height) / float(text_width));
	}

	for (i = j = k = 0; i < MAP_WIDTH * MAP_HEIGHT; ++i)
	{
		vertex_data[i][0] = GLfloat(j);
		vertex_data[i][2] = GLfloat(k);
		vertex_data[i][1] = map->data[k * MAP_WIDTH + (MAP_WIDTH - 1 - j) ] / 8.0f - 16;

		color_data[i][0] = 
		color_data[i][1] = 
		color_data[i][2] = map->data[k * MAP_WIDTH + (MAP_WIDTH - 1 - j) ] / 256.0f;

		texture_coords[i][0] = float(MAP_WIDTH - 1 - j) / TEXT_EDGE_LEN_X;
		texture_coords[i][1] = float(k) / TEXT_EDGE_LEN_Y;

		if (++j == MAP_WIDTH)
		{
			j = 0;
			++k;
		}
	}
	cons.Output("done\n");

	/* calculate d2 error values data */
	cons.Output("calculating d2 values... ");
	CalculateNormalized_d2(0, 0, MAP_WIDTH - 1, MAP_HEIGHT - 1);
	cons.Output("done\n");

	// select rendering function
	console->Separator();

	const GLubyte *gl_vendor, *gl_renderer, *gl_extensions, *gl_version;

	gl_vendor     = glGetString(GL_VENDOR);
	gl_extensions = glGetString(GL_EXTENSIONS);
	gl_renderer   = glGetString(GL_RENDERER);
	gl_version    = glGetString(GL_VERSION);
	cons.Output("OpenGL Information\nVendor: %s\nRenderer: %s\nVersion: %s\n", gl_vendor, gl_renderer, gl_version);
	cons.Output("Available OpenGL extensions: %s\n", gl_extensions);

	cons.Separator();
	QuadTreeRender = &QuadTreeRenderer::QuadTreeRender_Immediate;

	if (params.GET_INT_PARAM(PARAM_USE_ARB_EXT))
	{
		// query existing extensions 
		cons.Output("initializing ARB extensions\n");

		if (InitARBExtension(params) == GEZ_OK)
		{
			QuadTreeRender = &QuadTreeRenderer::QuadTreeRender_ARB;
			cons.Output("done, using ARB extension mode\n");
		}
		else
		{
			WARN("ARB extensions could not be initialized!");
			cons.Output("ARB extension init failed!\n");
		}
	}

	// in case of fallback to immediate inform user
	if (QuadTreeRender == &QuadTreeRenderer::QuadTreeRender_Immediate)
	{
		if (InitImmediateMode(params) == GEZ_OK)
		{
			cons.Output("Using OpenGL immediate drawing functions\n");
		}
		else
		{
			cons.Output("couldn't initialize immediate mode rendering functions\n");
			FAIL("no rendering function works here");
		}
	}

	/* set up OpenGL state */
	glClearColor(0.0, 0.0, 0.0, 0.0); 
	glClearDepth(1.0); 

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(params.GET_FLOAT_PARAM(PARAM_FIELD_OF_VIEW),
		params.GET_FLOAT_PARAM(PARAM_FIELD_OF_VIEW_ASPECT),
		params.GET_FLOAT_PARAM(PARAM_CLIP_PLANE_NEAR),
		params.GET_FLOAT_PARAM(PARAM_CLIP_PLANE_FAR));

	glPolygonMode(GL_FRONT, GL_FILL);
	glPolygonMode(GL_BACK, GL_FILL);

	glShadeModel(GL_SMOOTH);
	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);

	// texture state

	if (texture_mapping)
	{
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		ASSERT(glGetError() == GL_NO_ERROR, "texture error");
		glGenTextures(1, &texName);
		ASSERT(glGetError() == GL_NO_ERROR, "texture error");
		glBindTexture(GL_TEXTURE_2D, texName);
		ASSERT(glGetError() == GL_NO_ERROR, "texture error");

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		ASSERT(glGetError() == GL_NO_ERROR, "texture error");
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		ASSERT(glGetError() == GL_NO_ERROR, "texture error");
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, 
						GL_NEAREST);
		ASSERT(glGetError() == GL_NO_ERROR, "texture error");
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, 
						GL_NEAREST);
		ASSERT(glGetError() == GL_NO_ERROR, "texture error");
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, text_width, 
						text_height, 0, GL_RGB, GL_UNSIGNED_BYTE, 
						text_data);
		ASSERT(glGetError() == GL_NO_ERROR, "texture error");

		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
		ASSERT(glGetError() == GL_NO_ERROR, "texture error");
		glBindTexture(GL_TEXTURE_2D, texName);
		ASSERT(glGetError() == GL_NO_ERROR, "texture error");
	}


	// output information about rendering state
	console->ScrOutput("---=== QuadTree Renderer Initialized ===---\n");
	if (QuadTreeRender == &QuadTreeRenderer::QuadTreeRender_Immediate)
	{
		console->ScrOutput("Render Function: RenderImmediate()\n");
	}
	else if (QuadTreeRender == &QuadTreeRenderer::QuadTreeRender_Immediate_VertexArray)
	{
		console->ScrOutput("Render Function: RenderImmediate()\n");
		console->ScrOutput("using vertex_array_EXT\n");
	}
	else if (QuadTreeRender == &QuadTreeRenderer::QuadTreeRender_ARB)
	{
		console->ScrOutput("Render Function: Render_ARB_Extensions()\n");
	}
	else
	{
		FAIL("internal error --> invalid rendering algorithm\n");
	}
	if (multi_draw)
	{
		console->ScrOutput("using multidraw_EXT\n");
	}
	if (compiled_array)
	{
		console->ScrOutput("using compiled_vertex_array_EXT\n");
	}

	return GEZ_OK;
}

GLfloat QuadTreeRenderer::Calculate_d2_Recursion(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2)
{
	unsigned int x, y;
	float d;

	x = x1 + (x2 - x1) / 2;
	y = y1 + (y2 - y1) / 2;

	/* check whether to sub-divide or to calculate and propagate */
	if ((x1 + 1 >= x) || (x2 - 1 <= x))
	{
		/* lowest level of refinement --> calculate error values */
		float d1, d2, d3, d4, d5, d6;

		d1 = fabs(((map->data[y1 * MAP_WIDTH + x1] + map->data[y1 * MAP_WIDTH + x2]) * 0.5f) - map->data[y1 * MAP_WIDTH +  x]);
		d2 = fabs(((map->data[y1 * MAP_WIDTH + x2] + map->data[y2 * MAP_WIDTH + x2]) * 0.5f) - map->data[y  * MAP_WIDTH + x2]);
		d3 = fabs(((map->data[y2 * MAP_WIDTH + x1] + map->data[y2 * MAP_WIDTH + x2]) * 0.5f) - map->data[y2 * MAP_WIDTH +  x]);
		d4 = fabs(((map->data[y1 * MAP_WIDTH + x1] + map->data[y2 * MAP_WIDTH + x1]) * 0.5f) - map->data[y * MAP_WIDTH +  x1]);

		/* the diagonals */
		d5 = fabs(((map->data[y1 * MAP_WIDTH + x1] + map->data[y2 * MAP_WIDTH + x2]) * 0.5f) - map->data[y * MAP_WIDTH +  x]);
		d6 = fabs(((map->data[y2 * MAP_WIDTH + x1] + map->data[y1 * MAP_WIDTH + x2]) * 0.5f) - map->data[y * MAP_WIDTH +  x]);

		d = d1;
		if (d2 > d) d = d2;
		if (d3 > d) d = d3;
		if (d4 > d) d = d4;
		if (d5 > d) d = d5;
		if (d6 > d) d = d6;
	}
	else
	{
		/* sub-divide */
		float d1, d2, d3, d4;

		d1 = Calculate_d2_Recursion(x1,  y1,  x,  y);
		d2 = Calculate_d2_Recursion( x,  y1, x2,  y);
		d3 = Calculate_d2_Recursion( x,  y,  x2, y2);
		d4 = Calculate_d2_Recursion(x1,  y,  x,  y2);

		d = d1;
		if (d2 > d) d = d2;
		if (d3 > d) d = d3;
		if (d4 > d) d = d4;

		d_array[y * MAP_WIDTH + x] = d / ((float)x2 - x1); // store normalized error d2
	}

	return d; /* propagate back unnormalized error */
}

void QuadTreeRenderer::PreventCracks_d2(unsigned int edge_len)
{
	int x, y;
	int edge_len_half, edge_len_quarter;
	float d, K;
	float detail;

	edge_len_half = edge_len / 2;
	edge_len_quarter = edge_len_half / 2;

	detail = GetDetailLevel(); 
	K = detail / (2 * (detail - 1));

	if (edge_len > 4)
	{
		PreventCracks_d2(edge_len / 2);

		y = edge_len / 2;
		while (y < MAP_HEIGHT)
		{
			x = edge_len / 2;
			while (x < MAP_WIDTH)
			{
				d = d_array[MAP_WIDTH * y + x];

				if (x - edge_len_half - edge_len_quarter > 0)
				{
					if (y - edge_len_quarter > 0)
					{
						d = max(d, d_array[MAP_WIDTH * (y - edge_len_quarter) + x - edge_len_half - edge_len_quarter] * K);
					}

					if (y + edge_len_quarter < MAP_HEIGHT)
					{
						d = max(d, d_array[MAP_WIDTH * (y + edge_len_quarter) + x - edge_len_half - edge_len_quarter] * K);
					}
				}

				if (y - edge_len_half - edge_len_quarter > 0)
				{
					d = max(d, d_array[MAP_WIDTH * (y - edge_len_half - edge_len_quarter) + x - edge_len_quarter] * K);
					d = max(d, d_array[MAP_WIDTH * (y - edge_len_half - edge_len_quarter) + x + edge_len_quarter] * K);
				}

				if (y + edge_len_half + edge_len_quarter < MAP_HEIGHT)
				{
					d = max(d, d_array[MAP_WIDTH * (y + edge_len_half + edge_len_quarter) + x - edge_len_quarter] * K);
					d = max(d, d_array[MAP_WIDTH * (y + edge_len_half + edge_len_quarter) + x + edge_len_quarter] * K);
				}

				if (x + edge_len_half + edge_len_quarter < MAP_WIDTH)
				{
					if (y - edge_len_quarter > 0)
					{
						d = max(d, d_array[MAP_WIDTH * (y - edge_len_quarter) + x + edge_len_half + edge_len_quarter] * K);
					}

					if (y + edge_len_quarter < MAP_HEIGHT)
					{
						d = max(d, d_array[MAP_WIDTH * (y + edge_len_quarter) + x + edge_len_half + edge_len_quarter] * K);
					}
				}

				d_array[MAP_WIDTH * y + x] = d;

				x += edge_len;
			}
			y += edge_len;
		}
	}
}

void QuadTreeRenderer::CalculateNormalized_d2(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2)
{
	// calculate raw d values
	Calculate_d2_Recursion(x1, y1, x2, y2);

	// prevent cracks
	PreventCracks_d2(MAP_WIDTH);
}

void QuadTreeRenderer::QuadTreeTesselate(const unsigned int x1, const unsigned int y1, const unsigned int x2, const unsigned int y2, const Vector3f &vPos, bool cull)
{
	/* 
		global resolution criteria l / d < C
		l... distance eye -- vertex
		d... length of current node edge
		C... constant <=> overall level of detail

		L1-norm distance
		d = | x2 - x1 | + | y2 - y1 | + | z2 - z1 |

		f... (tresh hold) f = l / (d * C * MAX(c, 1))
		c... desired resolution
	*/

	float l;
	int d;
	float f;
	float d2;

	int x, y;

	d = x2 - x1;
	ASSERT(d > 0, "d is 0");

	x = x1 + (d >> 1); 
	y = y1 + ((y2 - y1) >> 1);

	l = fabs(vPos.x - x) + fabs(vPos.y - vertex_data[(MAP_WIDTH * y + x)][1] ) + fabs(vPos.z - y);
	ASSERT(l > 0, "l is 0");

	// perform frustum culling

	// terrain layout is  
	// 256/256 .... 0/256 
	// 256/0   .....0/0   
	
	Point p[4];
	
	if (cull)
	{
		Visibility vis;

		p[0].x = (float)x2; p[0].y = vertex_data[(MAP_WIDTH * y1 + x2)][1]; p[0].z = (float)y1; // near left
		p[1].x = (float)x1; p[1].y = vertex_data[(MAP_WIDTH * y1 + x1)][1]; p[1].z = (float)y1; // near right
		p[2].x = (float)x1; p[2].y = vertex_data[(MAP_WIDTH * y2 + x1)][1]; p[2].z = (float)y2; // far right
		p[3].x = (float)x2; p[3].y = vertex_data[(MAP_WIDTH * y2 + x2)][1]; p[3].z = (float)y2; // far left

		vis = isVisibleQuad(cull_planes, p);

		if (vis == FULL)
		{
			cull = false;
		}
		else if (vis == NONE)
		{
			quadtree_data[MAP_WIDTH * y + x] = NODE_CULLED;
			return;
		}
	}

	if (d > 3)
	{
		d2 = d_array[MAP_WIDTH * y + x]; // / ((float)d) not necessary; already done in calculated d2values()
		f = l / (d * C * ((c * d2) < 1 ? 1 : c * d2));

		if (f <= 1.0f)
		{
			/* subdivide this quad */
			quadtree_data[MAP_HEIGHT * y + x] = NODE_REFINE;

			QuadTreeTesselate(x1,  y1,  x,  y, vPos, cull);
			QuadTreeTesselate( x,  y1, x2,  y, vPos, cull);
			QuadTreeTesselate( x,  y,  x2, y2, vPos, cull);
			QuadTreeTesselate(x1,  y,  x,  y2, vPos, cull);
		}
		else
		{
			quadtree_data[MAP_HEIGHT * y + x] = NODE_RENDER;
		}
	}
	else
	{
		quadtree_data[MAP_HEIGHT * y + x] = NODE_RENDER;
		return;
	}
}



GEZERROR QuadTreeRenderer::Render(const Vector3f &vPos, const Vector3f &vRot)
{

	if (QuadTreeRender == NULL)
	{
		FAIL("QuadTreeRenderer is NULL!");
	}

	/* update viewport matrix */
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	pilotView(vPos, vRot);

	// get culling planes for frustum culling
	GetCullingPlanes(cull_planes);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glColor3f(1.0f, 1.0f, 1.0f); // unnecessary?

	if (texture)
		glEnable(GL_TEXTURE_2D);

	// show frustum culling process if enabled
	if (show_cull)
	{
		Vector3f vTemp, vTemp2;

		vTemp.x = vPos.x;
		vTemp.z = vPos.z - 50;
		vTemp.y = 50;

		vTemp2.x = 30;
		vTemp2.y = 180;
		vTemp2.z = 0;

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		pilotView(vTemp, vTemp2);
	}

	// calculate LOD quadtree 
	ZeroMemory(quadtree_data, MAP_WIDTH * MAP_HEIGHT); // necessary
	QuadTreeTesselate(0, 0, MAP_WIDTH - 1, MAP_HEIGHT - 1, vPos, frustum_culling);

	triangles_count = 0; vertices_count = 0; fan_current = 0;// collect statistical data
	(this->*QuadTreeRender)(0, 0, MAP_WIDTH - 1, MAP_HEIGHT - 1);

	glDisable(GL_TEXTURE_2D);

	glFlush();

	glDisable(GL_TEXTURE_2D);

	return GEZ_OK;
}



/* interface functions */
void QuadTreeRenderer::IncreaseDetailLevel()
{
	if (C < max_detail) 
	{
		console->ScrOutput("increasing global level of detail...\n", (int)C);
		++C;
		CalculateNormalized_d2(0, 0, MAP_WIDTH, MAP_HEIGHT);
		console->ScrOutput("global level of detail increased to %d\n", (int)C);
	}
	else
	{
		console->ScrOutput("maximum global detail reached\n");
	}
}


void QuadTreeRenderer::DecreaseDetailLevel()
{
	if (C > min_detail)
	{	
		console->ScrOutput("decreasing global level of detail...\n", (int)C);
		C--;
		CalculateNormalized_d2(0, 0, MAP_WIDTH, MAP_HEIGHT);
		console->ScrOutput("global level of detail decreased to %d\n", (int)C);
	}
	else
	{
		console->ScrOutput("minimum global detail reached\n");
	}
}

void QuadTreeRenderer::IncreaseLocalDetailLevel()
{
	if (c < 16)
	{
		++c;
		console->ScrOutput("local level of detail increased to %d\n", (int)c);
	}
	else
	{
		console->ScrOutput("maximum local detail reached\n");
	}
}


void QuadTreeRenderer::DecreaseLocalDetailLevel()
{
	if (c > 0)
	{	
		c--;
		console->ScrOutput("local level of detail decreased to %d\n", (int)c);
	}
	else
	{
		console->ScrOutput("minimum local detail reached\n");
	}
}

float QuadTreeRenderer::GetDetailLevel() const
{
	return C;
}

float QuadTreeRenderer::GetLocalDetailLevel() const
{
	return c;
}

QuadTreeRenderer::~QuadTreeRenderer()
{
  glDisableClientState(GL_VERTEX_ARRAY);

  CleanUp();

}

void QuadTreeRenderer::CleanUp()
{
#define FREEMEM(name)\
	if (name != NULL)\
	{\
		delete[] name;\
		name = NULL;\
	}

	FREEMEM(quadtree_data);
	FREEMEM(vertex_data);
	FREEMEM(color_data);
	FREEMEM(d_array);
	FREEMEM(texture_coords);

	FREEMEM(index_data);
	FREEMEM(multi_indexes);
	FREEMEM(multi_lengths);
	FREEMEM(multi_indexes_ARB);

	FREEMEM(text_data);

}
