/*

  this file deals with all operating system interaction

*/

#include "Application.h"

#include <windows.h> 

#include <stdio.h>

#include "Assert.h"
#include "main.h"

#include "Renderer.h"
#include "QuadTreeRenderer.h"
#include "Assert.h"

#ifndef NULL
#define NULL 0x0L
#endif

static Application *globalTempApp = NULL; // holds instance for WM_CREATE

Application *theApp = NULL;				// the global applation instance
Console     *g_console;					// the global console instance

Application::Application()
{
	initialized = false;

	theApp = this;

	hMouseEvent = NULL;
	ZeroMemory(bKey, 256);
}

GEZERROR Application::Initialize(HINSTANCE hThisInstance)
{
	WNDCLASSEX wincl;
	GEZERROR result;
	bool res;
	HRESULT  r;
	int window_style, width, heigth;

	ASSERT(initialized == false, "already initialized");
	if (initialized != false) return GEZ_ALREADYINITIALIZED;

	res = console.Initialize("log.txt");
	ASSERT(res, "console initialize failed");

	g_console = &console;

	console.Separator(true);
	console.Output("QuadTree Engine\n");
	console.Separator(true);
	console.Output("Reading configuration '%s'... ", CONFIG_FILE);

	/* load configuration data */
	if (!prefs.LoadConfig(CONFIG_FILE))
	{
		WARN("Couldn't open " CONFIG_FILE);
	}
	else
	{
        console.Output("done\n");
	}

	benchmark = prefs.GET_INT_PARAM(PARAM_BENCHMARK) != 0;
	frames_rendered = 0;

	hInstance = hThisInstance;

	width = prefs.GET_INT_PARAM(PARAM_WIDTH);
	heigth = prefs.GET_INT_PARAM(PARAM_HEIGHT);

	window_style = WS_SYSMENU | WS_DLGFRAME;

	/* The Window structure */
	wincl.hInstance = hThisInstance;
	wincl.lpszClassName = CLASSNAME;
	wincl.lpfnWndProc = WindowProcedure;      /* This function is called by windows */
	wincl.style = CS_DBLCLKS;                 /* Catch double-clicks */
	wincl.cbSize = sizeof (WNDCLASSEX);

	/* Use default icon and mouse-pointer */
	wincl.hIcon = LoadIcon (NULL, IDI_APPLICATION);
	wincl.hIconSm = LoadIcon (NULL, IDI_APPLICATION);
	wincl.hCursor = LoadCursor (NULL, IDC_ARROW);
	wincl.lpszMenuName = NULL;                 /* No menu */
	wincl.cbClsExtra = 0;                      /* No extra bytes after the window class */
	wincl.cbWndExtra = sizeof(long);                      /* structure or the window instance */
	/* Use Windows's default color as the background of the window */
	wincl.hbrBackground = (HBRUSH) COLOR_BACKGROUND;

	/* Register the window class, and if it fails quit the program */
	if (!RegisterClassEx (&wincl))
		return GEZ_UNKNOWN;

	/* load map data */
	if (camera.LoadMap(prefs.GET_STRING_PARAM(PARAM_MAP_NAME)) != GEZ_OK)
	{
		FAIL("can't load map");
	};

	camera.SetApp(this);
	if ((result = camera.Initialize()) != GEZ_OK)
	{
		return GEZ_UNKNOWN;
	}


	globalTempApp = this;

	if (prefs.GET_INT_PARAM(PARAM_FULLSCREEN))
	{
		//change display settings
		DEVMODE dmScreenSettings;
		ZeroMemory(&dmScreenSettings, sizeof(dmScreenSettings));
		dmScreenSettings.dmSize       = sizeof(dmScreenSettings);
		dmScreenSettings.dmFields     = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
		dmScreenSettings.dmPelsWidth  = width;
		dmScreenSettings.dmPelsHeight = heigth;
		dmScreenSettings.dmBitsPerPel = 16;

		if (ChangeDisplaySettings(&dmScreenSettings,CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
		{
			/* FIXME */
			MessageBox(NULL, "Couldn't set fullscreen --> window mode", "Error", MB_OK | MB_SYSTEMMODAL);
		}
		
		window_style = WS_POPUP;
	}

	/* The class is registered, let's create the program*/
	hwnd = CreateWindowEx (
			0,                   /* Extended possibilites for variation */
			CLASSNAME,           /* Classname */
			CAPTION,             /* Title Text */
			window_style,        /* default window */
			0,                   
			0,                   
			width,               /* The programs width */
			heigth,              /* and height in pixels */
			HWND_DESKTOP,        /* The window is a child-window to desktop */
			NULL,                /* No menu */
			hThisInstance,       /* Program Instance handler */
			NULL                 /* No Window Creation data */
			);

    /* handle input */
	console.Separator();
	console.Output("setting up input devices...");

	if (!benchmark)
	{
		/* initialize DirectX */
		r = DirectInput8Create(hInstance, DIRECTINPUT_VERSION, IID_IDirectInput8, (LPVOID *) &lpDInput, NULL);
		if (!SUCCEEDED(r))
		{
			console.Output("failed\n");
			return GEZ_UNKNOWN;
		}

		r = lpDInput->CreateDevice(GUID_SysMouse, &lpDIMouse, NULL);
		if (!SUCCEEDED(r))
		{
			console.Output("failed\n");
			return GEZ_UNKNOWN;
		}

		r = lpDIMouse->SetDataFormat(&c_dfDIMouse);
		if (!SUCCEEDED(r))
		{
			console.Output("failed\n");
			return GEZ_UNKNOWN;
		}

		r = lpDIMouse->SetCooperativeLevel(hwnd,
					DISCL_EXCLUSIVE | DISCL_FOREGROUND);
		if (!SUCCEEDED(r))
		{
			console.Output("failed\n");
			return GEZ_UNKNOWN;
		}

		hMouseEvent = CreateEvent(0, FALSE, 0, 0);

		if (hMouseEvent == NULL) {
			console.Output("failed\n");
			return GEZ_UNKNOWN;
		}

		r = lpDIMouse->SetEventNotification(hMouseEvent);

		if (FAILED(r)) {
			console.Output("failed\n");
			return GEZ_UNKNOWN;
		}

		#define DINPUT_BUFFERSIZE  16

		DIPROPDWORD dipdw =
		{
			// the header
			{
				sizeof(DIPROPDWORD),        // diph.dwSize
				sizeof(DIPROPHEADER),       // diph.dwHeaderSize
				0,                          // diph.dwObj
				DIPH_DEVICE,                // diph.dwHow
			},
			// the data
			DINPUT_BUFFERSIZE,              // dwData
		};

		r = lpDIMouse->SetProperty(DIPROP_BUFFERSIZE, &dipdw.diph);

		if (FAILED(r)) {
			console.Output("failed\n");
			return GEZ_UNKNOWN;
		}
	}

	console.Output("done\n");

	return GEZ_OK;
}


/*  This function is called by the Windows function DispatchMessage()  */
LRESULT CALLBACK Application::WindowProcedure (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    Application *app;
    Renderer *renderer;
    GEZERROR result;
	int i;
	BOOL res;

    switch (message)                  /* handle the messages */
    {

		case WM_ACTIVATE:
            app = (Application *) GetWindowLong(hwnd, GWL_USERDATA);

			app->bActive = wParam == WA_ACTIVE || wParam == WA_CLICKACTIVE;
			if (!app->benchmark)
			{
				if (app->bActive) app->lpDIMouse->Acquire();
			}
			break;

        case WM_KEYUP:
            app = (Application *) GetWindowLong(hwnd, GWL_USERDATA);
			if (!app->benchmark)
			{
				app = (Application *) GetWindowLong(hwnd, GWL_USERDATA);
				app->bKey[wParam] = 0;
			}
            break;

        case WM_KEYDOWN:
            app = (Application *) GetWindowLong(hwnd, GWL_USERDATA);
			if (!app->benchmark)
			{
				app = (Application *) GetWindowLong(hwnd, GWL_USERDATA);
				app->bKey[wParam] = 1;
			}

			if (wParam == VK_ESCAPE)
			{
				PostMessage(hwnd, WM_CLOSE, 0, 0);
			}

            break;
    
        case WM_CREATE:
        
            /* store pointer to this object-instance after window-data */
            ASSERT(globalTempApp != NULL, "globalTempApp is NULL");
            SetWindowLong(hwnd, GWL_USERDATA, (long)globalTempApp);

			app = globalTempApp;
            globalTempApp = NULL;

			app->hdc = GetWindowDC(hwnd);
			ASSERT(app->hdc != NULL, "hdc is null");

			/* get pixel format */
			ZeroMemory(&(app->pfd), sizeof(app->pfd));
			app->pfd.nSize      = sizeof(app->pfd);
			app->pfd.nVersion   = 1;
			app->pfd.dwFlags    = PFD_SUPPORT_OPENGL | PFD_DRAW_TO_WINDOW | PFD_DOUBLEBUFFER;
			app->pfd.iPixelType = PFD_TYPE_RGBA;
			app->pfd.iLayerType = PFD_MAIN_PLANE;
			app->pfd.cColorBits = 16;
			app->pfd.cDepthBits = 16;
			  
			// get best match for pixel format
			i = ChoosePixelFormat(app->hdc, &app->pfd);
			    
			ASSERT(i, "no pixel format match");

			if (i == 0)
			{
				return GEZ_UNKNOWN;
			}
			  
			// set pixel format
			app->console.Output("setting pixel format... ");
			SetPixelFormat(app->hdc, i, &(app->pfd));
			app->console.Output("done\n");

			app->hglrc = wglCreateContext(app->hdc);
			ASSERT(app->hglrc != NULL, "hglrc is null");
			if (app->hglrc == NULL) FAIL("NULL IS");

			res = wglMakeCurrent(app->hdc, app->hglrc);
			ASSERT(res, "wglmakecurrent failed");
			if (res == NULL) FAIL("NULL IS");

			if (app->console.init_gui() != true)
			{
				WARN("couldn't initialize GUI");
			}


            /* create renderer */
            renderer = new QuadTreeRenderer();
            ASSERT(renderer != NULL, "Renderer is NULL");
          
            if (renderer == NULL)
            {
              return false;
            }
          
            renderer->SetMap(app->camera.map);
            if ((result = renderer->Initialize(app->console, app->prefs) != GEZ_OK))
            {
              delete renderer;
              renderer = NULL;
            
              return false;
            }
 
            /* attach renderer to camera */
            app->camera.SetRenderer(renderer);

			app->threadID = GetCurrentThreadId();

            break;
            
        case WM_CLOSE:
            DestroyWindow(hwnd);
            break;    
      
        case WM_DESTROY:
            app = (Application *) GetWindowLong(hwnd, GWL_USERDATA);
			wglMakeCurrent(NULL, NULL);
			ReleaseDC(NULL, app->hdc);
			wglDeleteContext(app->hglrc);

            PostQuitMessage(0);       /* send a WM_QUIT to the message queue */
            break;
            
        default:                      /* for messages that we don't deal with */
            return DefWindowProc (hwnd, message, wParam, lParam);
    }

    return 0;
}

GEZERROR Application::Render()
{
	char buf[150];

	camera.Render();

	if (!benchmark)
	{
		wsprintf(buf, "%d fps, %d LLoD, %d GLoD, %d vertices, %d triangles, %d t-fans",
			(int)camera.fps,
			(int)((QuadTreeRenderer *)(camera.renderer))->c,
			(int)((QuadTreeRenderer *)(camera.renderer))->C,
			camera.renderer->vertices_count,
			camera.renderer->triangles_count,
			(int)((QuadTreeRenderer *)(camera.renderer))->fan_current);

		console.GuiResetYPos();
		console.ScrPrint(buf);
		console.ScrPrint("\n");
		console.ShowConsole(false);
	}
	else
	{
		++frames_rendered;
	}

	SwapBuffers(hdc);

	return GEZ_OK;
}


Application::~Application()
{
  Renderer *renderer;

  if (lpDIMouse)
  {
	  lpDIMouse->Release();
	  lpDIMouse = NULL;
  }

  if (lpDInput)
  {
	  lpDInput->Release();
	  lpDInput = NULL;
  }

  /* delete camera and the attached renderer */
  renderer = camera.GetRenderer();
  if (renderer != NULL)
  {
    delete renderer;
    renderer = NULL;
  }
}

void Application::MousePos(int x, int y)
{
	camera.RotateCam((float)x, (float)y); /* could be made faster by gem */
}
